/*
 * ***** BEGIN LICENSE BLOCK *****
 *
 * Copyright (c) 2011 ICSI Center for Internet Research (ICIR)
 * All rights reserved.
 *
 * See LICENSE.txt for license and terms of usage. 
 *
 * ***** END LICENSE BLOCK *****
 */

if (!fathomns) {
  var fathomns = {
    mod : {}
  };
}

Components.utils.import("resource://fathom/Logger.jsm",
    fathomns.mod);

/**
 * Provides functionality for the overlay. An instance of this class exists for
 * each tab/window.
 */
fathomns.overlay = {

  _initialized : false,
  _fathomService : null,

  // For things we can't do through the nsIFathom interface, use direct
  // access to the underlying JS object.
  _fathomServiceJSObject : null,

  _overlayId : null,
  _strbundle : null,

  toString : function() {
    return "[fathomns.overlay " + this._overlayId + "]";
  },

  /**
   * Initialize the object. This must be done after the DOM is loaded.
   */
  init : function() {
    try {
      if (this._initialized == false) {
        this._initialized = true;
        this._overlayId = (new Date()).getTime();

        this._fathomService = Components.classes["@icir.org/fathom-service;1"]
            .getService(Components.interfaces.nsIFathom);
        this._fathomServiceJSObject = this._fathomService.wrappedJSObject;

        this._strbundle = document.getElementById("fathomStrings");
      }
    } catch (e) {
      fathomns.mod.Logger.error("Fatal Error, " + e + ", stack was: " + e.stack);
      fathomns.mod.Logger.error("Unable to initialize fathomns.overlay.");
      throw e;
    }
  },

  onWindowClose : function(event) {
  },

  /**
   * Perform the actions required once the window has loaded.
   * 
   * @param {Event}
   *          event
   */
  onLoad : function(event) {
    try {

    } catch (e) {
      fathomns.mod.Logger.error("Fatal Error, " + e + ", stack was: " + e.stack);
      fathomns.mod.Logger.error("Unable to complete fathomns.overlay.onLoad actions.");
      throw e;
    }
  },

  /**
   * Perform the actions required once the DOM is loaded. This may be
   * called for more than just the page content DOM. 
   * 
   * @param {Event}
   *          event
   */
  _onDOMContentLoaded : function(document) {
    fathomns.mod.Logger.warning("_onDOMContentLoaded called.");
  }
};

// Initialize the fathomns.overlay object when the window DOM is loaded.
addEventListener("DOMContentLoaded", function(event) {
      fathomns.overlay.init();
    }, false);

// Event handler for when the window is closed. We listen for "unload" rather
// than "close" because "close" will fire when a print preview opened from this
// window is closed.
addEventListener("unload", function(event) {
      fathomns.overlay.onWindowClose(event);
    }, false);

// Registers event handlers for documents loaded in the window.
addEventListener("load", function(event) {
      fathomns.overlay.onLoad(event);
    }, false);

/* code for handling overlay events */
var pref = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefBranch);

var fathom_toggleHelpers = {

	getCurrentStatus : function() {
		return pref.getBoolPref("extensions.fathom.status");
	},

	setCurrentStatus : function(val){
		pref.setBoolPref("extensions.fathom.status", val);
	},

	toggleCurrentStatus : function(prefix) {
		if (this.getCurrentStatus()) {
			item = document.getElementById("fathom-icon");
			if(item)
				item.setAttribute("image", "chrome://fathom/content/icons/off.png");

			item = document.getElementById("toolbar-fathom-button");
			if(item)
				item.setAttribute("image", "chrome://fathom/content/icons/off.png");

			item = document.getElementById(prefix + "fathom-disable");
			if(item)
				item.setAttribute("checked", "true");

			this.setCurrentStatus(false);
		} else {
			item = document.getElementById("fathom-icon");
			if(item)
				item.setAttribute("image", "chrome://fathom/content/icons/on.png");

			item = document.getElementById("toolbar-fathom-button");
			if(item)
				item.setAttribute("image", "chrome://fathom/content/icons/on.png");

			item = document.getElementById(prefix + "fathom-disable");
			if(item)
				item.setAttribute("checked", "false");

			this.setCurrentStatus(true);
		}
	}
}

var fathom_helpers = {

	onLoad: function() {
		var parent = document.getElementById("status-bar");
		var entry = null, item = null;
		if (!fathom_toggleHelpers.getCurrentStatus()) {
			// set icon-image to off
			item = document.getElementById("fathom-icon");
			if(item)
				item.setAttribute("image", "chrome://fathom/content/icons/off.png");

			item = document.getElementById("toolbar-fathom-button");
			if(item)
				item.setAttribute("image", "chrome://fathom/content/icons/off.png");
		}
	},
	
	status: function(){
		return fathom_toggleHelpers.getCurrentStatus();
	},
	
	onMenuItemCommand: function(e, type, prefix) {
		var item = null;
		switch(type) {
			case 'toggle':
				fathom_toggleHelpers.toggleCurrentStatus(prefix);
				break;

			case 'bugreport':
				var loc = gBrowser.contentDocument.location;
				var xulRuntime = Components.classes["@mozilla.org/xre/app-info;1"].getService(Components.interfaces.nsIXULRuntime);
				var appInfo = Components.classes["@mozilla.org/xre/app-info;1"].getService(Components.interfaces.nsIXULAppInfo);;
				var platform = appInfo.vendor + " " + appInfo.name + " " + appInfo.version + " / " + xulRuntime.OS;
				gBrowser.loadURI('mailto:fathom-bug@icsi.berkeley.edu?subject=Bug report: Build #' + buildInfo + ', Browser/OS [' + platform + '], URL [' + loc + ']');
				break;
				
			case 'about':
				window.open("chrome://fathom/content/about.xul", "About", "chrome,centerscreen");
				break;
				
			case 'status':
				fathom_toggleHelpers.toggleMessagePanel(prefix);
				break;
				
			default:
				break;
		}
	},
	
	onPopUpShowing: function(e, prefix) {
		var item = null;
		// set status info
		if (fathom_toggleHelpers.getCurrentStatus()) {
			item = document.getElementById(prefix + "fathom-disable");
			if(item)
				item.setAttribute("checked", "false");
		} else {
			item = document.getElementById(prefix + "fathom-disable");
			if(item)
				item.setAttribute("checked", "true");
		}
		// set version info
		var version = document.getElementById(prefix + "fathom-version");
		if(version)
			version.label = "Build : " + fathomBuildInfo;
	},

	onIconClick: function(e) {
		// toggle status on middle button click
		if (e.button == 1) {
			fathom_toggleHelpers.toggleCurrentStatus();
		}
	}
};

window.addEventListener("load", function(e) { fathom_helpers.onLoad(e); }, false);

